# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import logging

from cli.model.atomic import CLIAtomicProfile, CLIAtomicCommandGroup, CLIAtomicCommandGroupRegisterInfo, \
    CLIAtomicCommand, CLIAtomicCommandRegisterInfo, CLISpecsResource, CLICommandGroupHelp, CLICommandHelp, \
    CLICommandExample, CLIAtomicClient
from command.controller.cfg_reader import CfgReader
from command.controller.specs_manager import AAZSpecsManager
from command.model.configuration import CMDHttpOperation, CMDCommand, CMDArgGroup, CMDObjectOutput, \
    CMDHttpResponseJsonBody, CMDObjectSchemaBase
from swagger.utils.tools import swagger_resource_path_to_resource_id
from utils.stage import AAZStageEnum
from utils.exceptions import ResourceNotFind
from utils.plane import PlaneEnum
from utils.case import to_camel_case, to_snake_case

logger = logging.getLogger('backend')


class AzAtomicProfileBuilder:

    def __init__(self, mod_name, by_patch=False):
        self._mod_name = mod_name
        self._aaz_spec_manager = AAZSpecsManager()
        self._by_patch = by_patch

    def __call__(self, view_profile):
        profile = CLIAtomicProfile()
        profile.name = view_profile.name
        if view_profile.command_groups:
            cmd_groups = {}
            cmd_clients = {}
            for name, view_cmd_group in view_profile.command_groups.items():
                cmd_group, clients = self._build_command_group(view_cmd_group)
                cmd_groups[name] = cmd_group
                cmd_clients.update(clients)
            profile.command_groups = cmd_groups
            for client in cmd_clients.values():
                profile.add_client(client)
        return profile

    def _build_command_group(self, view_command_group):
        command_group = self._build_command_group_from_aaz(*view_command_group.names)
        stages = set()
        cmd_clients = {}

        if view_command_group.commands:
            # always load cfg for full generation
            load_cfg = not self._by_patch
            cmds = {}
            if not load_cfg:
                for name, view_cmd in view_command_group.commands.items():
                    if view_cmd.modified:
                        # when one or more command modified, load the cfg of all the commands in this command group
                        # BTW, sub command groups are not included
                        load_cfg = True
                        break
            for name, view_cmd in view_command_group.commands.items():
                cmd, client = self._build_command(view_cmd, load_cfg)
                if cmd.register_info is not None:
                    stages.add(cmd.register_info.stage)
                cmds[name] = cmd
                cmd_clients[(client.plane, client.name)] = client

            command_group.commands = cmds
            if load_cfg:
                command_group.wait_command = self._complete_command_wait_info(command_group)
            elif view_command_group.wait_command and len(stages):
                # keep wait file not changed.
                command_group.wait_command = CLIAtomicCommand()
                command_group.wait_command.names = [*view_command_group.wait_command.names]
                command_group.wait_command.register_info = CLIAtomicCommandRegisterInfo()

        if view_command_group.command_groups:
            cmd_groups = {}
            for name, view_cmd_group in view_command_group.command_groups.items():
                cmd_group, clients = self._build_command_group(view_cmd_group)
                if cmd_group.register_info is not None:
                    stages.add(cmd_group.register_info.stage)
                cmd_groups[name] = cmd_group
                cmd_clients.update(clients)
            command_group.command_groups = cmd_groups

        if AAZStageEnum.Stable in stages:
            command_group.register_info.stage = AAZStageEnum.Stable
        elif AAZStageEnum.Preview in stages:
            command_group.register_info.stage = AAZStageEnum.Preview
        elif AAZStageEnum.Experimental in stages:
            command_group.register_info.stage = AAZStageEnum.Experimental
        elif not stages:
            command_group.register_info = None
        else:
            raise NotImplementedError()

        return command_group, cmd_clients

    def _build_command(self, view_command, load_cfg):
        command = self._build_command_from_aaz(*view_command.names, version_name=view_command.version, load_cfg=load_cfg)
        if not view_command.registered:
            command.register_info = None
        client = self._build_client_from_aaz(plane=command.resources[0].plane)
        return command, client

    def _build_command_group_from_aaz(self, *names):
        aaz_cg = self._aaz_spec_manager.find_command_group(*names)
        if not aaz_cg:
            raise ResourceNotFind("Command group '{}' not exist in AAZ".format(' '.join(names)))
        command_group = CLIAtomicCommandGroup()
        command_group.names = [*names]
        command_group.help = CLICommandGroupHelp()
        command_group.help.short = aaz_cg.help.short
        if aaz_cg.help.lines:
            command_group.help.long = '\n'.join(aaz_cg.help.lines)
        command_group.register_info = CLIAtomicCommandGroupRegisterInfo({
            "stage": AAZStageEnum.Stable
        })
        return command_group

    def _build_command_from_aaz(self, *names, version_name, load_cfg=True):
        aaz_cmd = self._aaz_spec_manager.find_command(*names)
        if not aaz_cmd:
            raise ResourceNotFind("Command '{}' not exist in AAZ".format(' '.join(names)))
        version = None
        for v in (aaz_cmd.versions or []):
            if v.name == version_name:
                version = v
                break
        if not version:
            raise ResourceNotFind("Version '{}' of command '{}' not exist in AAZ".format(version_name, ' '.join(names)))

        command = CLIAtomicCommand()
        command.names = [*names]
        command.help = CLICommandHelp()
        command.help.short = aaz_cmd.help.short
        if aaz_cmd.help.lines:
            command.help.long = '\n'.join(aaz_cmd.help.lines)

        if version.examples:
            command.help.examples = [CLICommandExample(e.to_primitive()) for e in version.examples]
        command.version = version.name
        command.stage = version.stage or AAZStageEnum.Stable
        command.resources = [CLISpecsResource(r.to_primitive()) for r in version.resources]
        command.register_info = CLIAtomicCommandRegisterInfo({
            "stage": command.stage,
        })

        if load_cfg:
            # load cfg file, which will generate the command in code
            cfg_reader = self._aaz_spec_manager.load_resource_cfg_reader_by_command_with_version(
                aaz_cmd, version=version)
            cmd_cfg = cfg_reader.find_command(*names)
            assert cmd_cfg is not None, f"command model miss in AAZ: '{' '.join(names)}'"

            command.cfg = cmd_cfg
            command.register_info.confirmation = cmd_cfg.confirmation
        return command

    def _build_client_from_aaz(self, plane):
        client = CLIAtomicClient({
            "plane": plane,
            "name": PlaneEnum.http_client(plane)
        })
        if plane in PlaneEnum._config:
            # use the clients registered in azure/cli/core/aaz/_client.py
            client.registered_name = client.name
            client.cls_name = client.name
        else:
            # generate client based on client config
            cfg_reader = self._aaz_spec_manager.load_client_cfg_reader(plane)
            assert cfg_reader, "Missing Client config for '" + plane + "' plane."
            client.cfg = cfg_reader.cfg
            scope = PlaneEnum.get_data_plane_scope(plane) or plane
            client.cls_name = to_camel_case(f"AAZ {scope.replace('.', ' ')} {client.name}")
            client.registered_name = f'{client.cls_name}_{to_snake_case(self._mod_name)}'  # for example: AAZAzureCodesigningDataPlaneClient_network
        return client

    @classmethod
    def _complete_command_wait_info(cls, command_group):
        assert command_group.commands
        wait_cmd_rids = {}
        for command in command_group.commands.values():
            lro_list = []
            for operation in command.cfg.operations:
                if isinstance(operation, CMDHttpOperation):
                    if operation.long_running:
                        lro_list.append(operation)

            if len(lro_list) == 1:
                # support no wait if there are only one long running operation
                # not support multiple long running operations
                command.support_no_wait = True
                if command.register_info is not None:
                    # command is registered
                    rid = swagger_resource_path_to_resource_id(lro_list[0].http.path)
                    if rid not in wait_cmd_rids:
                        wait_cmd_rids[rid] = {
                            "methods": set()
                        }

        if not wait_cmd_rids:
            return

        # build wait command
        for command in command_group.commands.values():
            for operation in command.cfg.operations:
                # find get operations for wait command
                if not isinstance(operation, CMDHttpOperation):
                    continue
                rid = swagger_resource_path_to_resource_id(operation.http.path)
                if rid not in wait_cmd_rids:
                    continue
                if operation.http.request.method != 'get':
                    wait_cmd_rids[rid]['methods'].add(operation.http.request.method)
                    continue
                if 'get_op' in wait_cmd_rids[rid]:
                    continue

                # verify operation response has provisioning state field
                if not cls._has_provisioning_state(operation):
                    continue

                wait_cmd_rids[rid]['get_op'] = operation.__class__(operation.to_primitive())
                wait_cmd_rids[rid]['args'] = {}
                for resource in command.resources:
                    if rid == resource.id:
                        wait_cmd_rids[rid]['resource'] = resource.__class__(resource.to_primitive())

                params = []
                if operation.http.request.path and operation.http.request.path.params:
                    params += operation.http.request.path.params
                if operation.http.request.query and operation.http.request.query.params:
                    params += operation.http.request.query.params
                if operation.http.request.header and operation.http.request.header.params:
                    params += operation.http.request.header.params
                for param in params:
                    if not param.arg:
                        continue
                    assert param.arg.startswith("$"), f"Not support path arg name: '{param.arg}'"
                    arg, arg_idx = CfgReader.find_arg_in_command_by_var(
                        command=command.cfg,
                        arg_var=param.arg
                    )
                    assert arg is not None
                    wait_cmd_rids[rid]['args'][arg_idx] = arg.__class__(arg.to_primitive())

        for rid, value in [*wait_cmd_rids.items()]:
            if "get_op" not in value:
                logger.error(f'Failed to support wait command for resource: '
                             f'Get operation with provisioning state property does not exist: {rid}')
                del wait_cmd_rids[rid]

        if not wait_cmd_rids:
            return

        if len(wait_cmd_rids) > 1:
            # Not support to generate wait command for multiple resources
            logger.error(f'A wait command cannot apply on multiple resources')
            return

        wait_cmd_info = [*wait_cmd_rids.values()][0]

        wait_command = CLIAtomicCommand()
        wait_command.names = [*command_group.names, "wait"]
        wait_command.help = CLICommandHelp()
        wait_command.help.short = "Place the CLI in a waiting state until a condition is met."
        wait_command.register_info = CLIAtomicCommandRegisterInfo()
        wait_command.resources = [wait_cmd_info['resource']]
        wait_command.cfg = cfg = CMDCommand()
        cfg.name = "wait"
        cfg.version = "undefined"

        arg_group = CMDArgGroup()
        cfg.arg_groups = [arg_group]
        arg_group.name = ""
        arg_group.args = [
            *wait_cmd_info['args'].values()
        ]
        get_op = wait_cmd_info['get_op']
        cfg.operations = [get_op]

        output = CMDObjectOutput()
        for response in get_op.http.responses:
            if response.is_error:
                continue
            if not isinstance(response.body, CMDHttpResponseJsonBody):
                continue
            if response.body.json.var:
                output.ref = response.body.json.var
                break
        if not output.ref:
            raise ValueError("Output ref is empty")
        output.client_flatten = False
        cfg.outputs = [output]
        return wait_command

    @staticmethod
    def _has_provisioning_state(get_op):
        for response in get_op.http.responses:
            if response.is_error:
                continue
            if not isinstance(response.body, CMDHttpResponseJsonBody):
                continue
            if not isinstance(response.body.json.schema, CMDObjectSchemaBase):
                continue
            schema = response.body.json.schema
            if schema.props:
                for prop in schema.props:
                    if prop.name.lower() in ("provisioning_state", "provisioningstate"):
                        return True
                    if prop.name.lower() == "properties" and \
                            isinstance(prop, CMDObjectSchemaBase) and prop.props:
                        for sub_prop in prop.props:
                            if sub_prop.name.lower() in ("provisioning_state", "provisioningstate"):
                                return True
                            if sub_prop.name.lower() in ("additional_properties", "additionalproperties") and \
                                    isinstance(sub_prop, CMDObjectSchemaBase) and sub_prop.props:
                                for p in sub_prop.props:
                                    if p.name.lower() in ("provisioning_state", "provisioningstate"):
                                        return True
        return False



import click
import logging
from flask import Blueprint
import sys

from utils.config import Config

logger = logging.getLogger('backend')

bp = Blueprint('cli-cmds', __name__, url_prefix='/CLI/CMDs', cli_group="cli")
bp.cli.short_help = "Manage aaz commands in azure-cli and azure-cli-extensions."


@bp.cli.command("regenerate", short_help="Regenerate aaz commands from command models in azure-cli/azure-cli-extensions")
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--cli-path", '-c',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    callback=Config.validate_and_setup_cli_path,
    help="The local path of azure-cli repo. Only required when generate code to azure-cli repo."
)
@click.option(
    "--cli-extension-path", '-e',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    callback=Config.validate_and_setup_cli_extension_path,
    help="The local path of azure-cli-extension repo. Only required when generate code to azure-cli-extension repo."
)
@click.option(
    "--extension-or-module-name", '--name',
    required=True,
    help="Name of the module in azure-cli or the extension in azure-cli-extensions"
)
def regenerate_code(extension_or_module_name, cli_path=None, cli_extension_path=None):
    from utils.config import Config
    from utils.exceptions import InvalidAPIUsage
    from cli.controller.az_module_manager import AzExtensionManager, AzMainManager
    if not cli_path and not cli_extension_path:
        logger.error("Please provide `--cli-path` or `--cli-extension-path`")
        sys.exit(1)
    if cli_path and cli_extension_path:
        logger.error("Please don't provide `--cli-path` and `--cli-extension-path`")
        sys.exit(1)

    try:
        if cli_path is not None:
            assert Config.CLI_PATH is not None
            manager = AzMainManager()
        else:
            assert cli_extension_path is not None
            assert Config.CLI_EXTENSION_PATH is not None
            manager = AzExtensionManager()

        if not manager.has_module(extension_or_module_name):
            # module = manager.create_new_mod(extension_or_module_name)
            raise ValueError(f"Cannot find module or extension `{extension_or_module_name}`")
        logger.info(f"Load module `{extension_or_module_name}`")
        module = manager.load_module(extension_or_module_name)
        logger.info(f"Regenerate module `{extension_or_module_name}`")
        manager.update_module(extension_or_module_name, module.profiles)
    except InvalidAPIUsage as err:
        logger.error(err)
        sys.exit(1)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)


@bp.cli.command("generate-by-swagger-tag", short_help="Generate aaz commands from command models in azure-cli/azure-cli-extensions selected by swagger tags.")
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--cli-path", '-c',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    callback=Config.validate_and_setup_cli_path,
    help="The local path of azure-cli repo. Only required when generate code to azure-cli repo."
)
@click.option(
    "--cli-extension-path", '-e',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    callback=Config.validate_and_setup_cli_extension_path,
    help="The local path of azure-cli-extension repo. Only required when generate code to azure-cli-extension repo."
)
@click.option(
    "--extension-or-module-name", '--name',
    required=True,
    help="Name of the module in azure-cli or the extension in azure-cli-extensions"
)
@click.option(
    "--swagger-module-path", "--sm",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_MODULE_PATH,
    required=not Config.SWAGGER_MODULE_PATH,
    callback=Config.validate_and_setup_swagger_module_path,
    expose_value=False,
    help="The local path of swagger module."
)
@click.option(
    "--resource-provider", "--rp",
    default=Config.DEFAULT_RESOURCE_PROVIDER,
    required=not Config.DEFAULT_RESOURCE_PROVIDER,
    callback=Config.validate_and_setup_default_resource_provider,
    expose_value=False,
    help="The resource provider name."
)
@click.option(
    "--swagger-tag", "--tag",
    required=True,
    help="Swagger tag with input files."
)
@click.option(
    "--profile",
    required=True,
    type=click.Choice(Config.CLI_PROFILES),
    default=Config.CLI_DEFAULT_PROFILE,
)
def generate_by_swagger_tag(profile, swagger_tag, extension_or_module_name, cli_path=None, cli_extension_path=None):
    from utils.config import Config
    from utils.exceptions import InvalidAPIUsage
    from cli.controller.az_module_manager import AzExtensionManager, AzMainManager
    from swagger.controller.specs_manager import SwaggerSpecsManager
    from command.controller.specs_manager import AAZSpecsManager
    if not Config.DEFAULT_SWAGGER_MODULE:
        Config.DEFAULT_SWAGGER_MODULE = "__MODULE__"

    try:
        swagger_specs = SwaggerSpecsManager()
        aaz_specs = AAZSpecsManager()
        module_manager = swagger_specs.get_module_manager(Config.DEFAULT_PLANE, Config.DEFAULT_SWAGGER_MODULE)
        rp = module_manager.get_openapi_resource_provider(Config.DEFAULT_RESOURCE_PROVIDER)

        resource_map = rp.get_resource_map_by_tag(swagger_tag)
        if not resource_map:
            raise InvalidAPIUsage(f"Tag `{swagger_tag}` is not exist")

        commands_map = {}
        for resource_id, version_map in resource_map.items():
            v_list = [v for v in version_map]
            if len(v_list) > 1:
                raise InvalidAPIUsage(f"Tag `{swagger_tag}` contains multiple api versions of one resource", payload={
                    "Resource": resource_id,
                    "versions": v_list,
                })

            v = v_list[0]
            # TODO: handle plane here
            cfg_reader = aaz_specs.load_resource_cfg_reader(Config.DEFAULT_PLANE, resource_id, v)
            if not cfg_reader:
                logger.error(f"Command models not exist in aaz for resource: {resource_id} version: {v}")
                continue
            for cmd_names, command in cfg_reader.iter_commands():
                key = tuple(cmd_names)
                if key in commands_map and commands_map[key] != command.version:
                    raise ValueError(f"Multi version contained for command: {''.join(cmd_names)} versions: {commands_map[key]}, {command.version}")
                commands_map[key] = command.version

        profile = _build_profile(profile, commands_map)

        if cli_path is not None:
            assert Config.CLI_PATH is not None
            manager = AzMainManager()
        else:
            assert cli_extension_path is not None
            assert Config.CLI_EXTENSION_PATH is not None
            manager = AzExtensionManager()

        if not manager.has_module(extension_or_module_name):
            logger.info(f"Create cli module `{extension_or_module_name}`")
            manager.create_new_mod(extension_or_module_name)

        logger.info(f"Load cli module `{extension_or_module_name}`")
        module = manager.load_module(extension_or_module_name)

        module.profiles[profile.name] = profile
        logger.info(f"Regenerate module `{extension_or_module_name}`")
        manager.update_module(extension_or_module_name, module.profiles)

    except InvalidAPIUsage as err:
        logger.error(err)
        sys.exit(1)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)


def _build_profile(profile_name, commands_map):
    from cli.model.view import CLIViewProfile, CLIViewCommand, CLIViewCommandGroup
    profile = CLIViewProfile({
        "name": profile_name,
        "commandGroups": {},
    })
    command_group_map = {tuple(): profile}
    for cmd_names, version in commands_map.items():
        group_names = cmd_names[:-1]
        if group_names not in command_group_map:
            group = CLIViewCommandGroup({
                "names": list(group_names),
                "commandGroups": {},
                "commands": {}
            })
        else:
            group = command_group_map[group_names]
        group.commands[cmd_names[-1]] = CLIViewCommand({
            "names": list(cmd_names),
            "version": version,
            "registered": True
        })

        while group_names not in command_group_map:
            command_group_map[group_names] = group
            parent_group_names = group_names[:-1]
            if parent_group_names not in command_group_map:
                parent_group = CLIViewCommandGroup({
                    "names": list(parent_group_names),
                    "commandGroups": {},
                    "commands": {}
                })
            else:
                parent_group = command_group_map[parent_group_names]
            parent_group.command_groups[group_names[-1]] = group

            group = parent_group
            group_names = parent_group_names

    return profile



from schematics.models import Model
from schematics.types import BaseType, StringType, ModelType, DictType, PolyModelType

from command.model.configuration import CMDHttpResponse, CMDHttpResponseHeader, CMDHttpResponseJsonBody, \
    CMDObjectSchemaBase, CMDArraySchemaBase, CMDHttpResponseHeaderItem, CMDClsSchemaBase, CMDResponseJson, \
    CMDBooleanSchemaBase, CMDStringSchemaBase, CMDFloatSchemaBase, CMDIntegerSchemaBase
from swagger.model.schema.fields import MutabilityEnum
from swagger.utils import exceptions
from utils.error_format import AAZErrorFormatEnum
from .example_item import XmsExamplesField
from .fields import XmsErrorResponseField, XNullableField
from .header import Header
from .reference import Linkable
from .schema import Schema, ReferenceSchema, schema_and_reference_schema_claim_function


class Response(Model, Linkable):
    """Describes a single response from an API Operation."""

    description = StringType(
        required=True)  # A short description of the response. GFM syntax can be used for rich text representation.
    schema = PolyModelType(
        [ModelType(Schema), ModelType(ReferenceSchema)],
        claim_function=schema_and_reference_schema_claim_function
    )  # A definition of the response structure. It can be a primitive, an array or an object. If this field does not exist, it means no content is returned as part of the response. As an extension to the Schema Object, its root type value may also be "file". This SHOULD be accompanied by a relevant produces mime-type.
    headers = DictType(ModelType(Header))  # A list of headers that are sent with the response.
    examples = DictType(BaseType())  # TODO:

    x_ms_examples = XmsExamplesField()  # TODO:
    x_ms_error_response = XmsErrorResponseField()

    x_nullable = XNullableField(
        default=False)  # TODO: # when true, specifies that null is a valid value for the associated schema

    def link(self, swagger_loader, *traces, **kwargs):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        if self.schema is not None:
            self.schema.link(swagger_loader, *self.traces, 'schema')

        # assign resource id template to the schema and it's ref instance
        resource_id_template = kwargs.get('resource_id_template', None)
        if resource_id_template and self.schema:
            if isinstance(self.schema, Schema):
                self.schema.resource_id_templates.add(resource_id_template)
            if self.schema.ref_instance and isinstance(self.schema.ref_instance, Schema):
                self.schema.ref_instance.resource_id_templates.add(resource_id_template)

        # TODO: add support for examples and x_ms_examples

    @classmethod
    def _claim_polymorphic(cls, data):
        if isinstance(data, dict):
            return 'description' in data
        elif isinstance(data, Response):
            return True
        return False

    def to_cmd(self, builder, is_error=False, status_codes=None, **kwargs):
        is_error = self.x_ms_error_response or is_error
        response = CMDHttpResponse()
        response.is_error = is_error
        response.status_codes = sorted(status_codes or [])
        builder.setup_description(response, self)

        if self.headers:
            response.header = CMDHttpResponseHeader()
            response.header.items = []
            for name in sorted(self.headers):
                header = CMDHttpResponseHeaderItem()
                header.name = name
                response.header.items.append(header)

        if self.schema:
            v = builder(
                self.schema,
                mutability=MutabilityEnum.Read,
                in_base=True,
                support_cls_schema=True,
            )
            if v.frozen:
                raise exceptions.InvalidSwaggerValueError(
                    msg="Invalid Response Schema. It's None.",
                    key=self.traces,
                )

            if isinstance(v, (
                    CMDStringSchemaBase,
                    CMDObjectSchemaBase,
                    CMDArraySchemaBase,
                    CMDBooleanSchemaBase,
                    CMDFloatSchemaBase,
                    CMDIntegerSchemaBase,
                    CMDClsSchemaBase,
            )):
                model = CMDResponseJson()
                if is_error:
                    error_format = AAZErrorFormatEnum.classify_error_format(builder, v)
                    if error_format:
                        err_schema = CMDClsSchemaBase()
                        err_schema.read_only = v.read_only
                        err_schema.frozen = v.frozen
                        err_schema._type = f"@{error_format}"
                        v = err_schema
                    else:
                        raise exceptions.InvalidSwaggerValueError(
                            msg="Error response schema is not supported yet.",
                            key=self.traces,
                        )
                model.schema = v
            else:
                raise exceptions.InvalidSwaggerValueError(
                    msg="Invalid Response type",
                    key=self.traces,
                    value=v.type
                )

            if isinstance(model, CMDResponseJson):
                response.body = CMDHttpResponseJsonBody()
                response.body.json = model
            else:
                raise NotImplementedError()

        return response



from cli.model.common._fields import CLICommandNameField
from schematics.models import Model
from schematics.types import ModelType, ListType, DictType
from utils.stage import AAZStageField

from ._command import CLIAtomicCommand
from ._help import CLICommandGroupHelp


class CLIAtomicCommandGroupRegisterInfo(Model):
    stage = AAZStageField(required=True)  # the stage of command group used in code
    # TODO: add support for deprecate_info


class CLIAtomicCommandGroup(Model):
    names = ListType(field=CLICommandNameField(), min_size=1, required=True)  # full name of a command group
    help = ModelType(CLICommandGroupHelp, required=True)
    register_info = ModelType(
        CLIAtomicCommandGroupRegisterInfo,
        required=False,
        serialized_name="registerInfo",
        deserialize_from="registerInfo"
    )  # register info in command group table. If it's not registered in command group table, this field is None

    command_groups = DictType(
        field=ModelType("CLIAtomicCommandGroup"),
        serialized_name="commandGroups",
        deserialize_from="commandGroups"
    )
    commands = DictType(
        field=ModelType(CLIAtomicCommand)
    )
    wait_command = ModelType(CLIAtomicCommand)

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, ModelType


class XmsParameterGrouping(Model):
    """
    By default operation parameters are generated in the client as method arguments. This behavior can sometimes be undesirable when the number of parameters is high. x-ms-parameter-grouping extension is used to group multiple primitive parameters into a composite type to improve the API.
    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-parameter-grouping
    """

    name = StringType()  # When set, specifies the name for the composite type.
    postfix = StringType()  # Alternative to name parameter. If specified the name of the composite type will be generated as follows {MethodGroup}{Method}{Postfix}


class XmsParameterGroupingField(ModelType):

    def __init__(self, **kwargs):
        super(XmsParameterGroupingField, self).__init__(
            XmsParameterGrouping,
            serialized_name="x-ms-parameter-grouping",
            deserialize_from="x-ms-parameter-grouping",
            **kwargs
        )



from schematics.types import StringType
from .items import Items
from .fields import XmsClientNameField, XmsHeaderCollectionPrefix


class Header(Items):
    description = StringType()     # A short description of the header.

    x_ms_client_name = XmsClientNameField()

    # specific properties, will not support
    x_ms_header_collection_prefix = XmsHeaderCollectionPrefix()  # only used in Storage Data plane



import re

import inflect
from utils.case import to_camel_case

from ._arg import CMDArg, CMDArgBase, CMDArgumentHelp, CMDArgEnum, CMDArgDefault, CMDBooleanArgBase, \
    CMDArgBlank, CMDObjectArgAdditionalProperties, CMDResourceLocationArgBase, CMDClsArgBase, CMDPasswordArgPromptInput
from ._format import CMDFormat
from ._schema import CMDObjectSchema, CMDSchema, CMDSchemaBase, CMDObjectSchemaBase, CMDObjectSchemaDiscriminator, \
    CMDArraySchemaBase, CMDObjectSchemaAdditionalProperties, CMDResourceIdSchema, CMDBooleanSchemaBase, \
    CMDResourceLocationSchemaBase, CMDPasswordSchema


class CMDArgBuilder:
    _inflect_engine = inflect.engine()

    @classmethod
    def new_builder(cls, schema, parent=None, var_prefix=None, ref_args=None, ref_arg=None, is_update_action=False):
        if var_prefix is None:
            if parent is None or parent._arg_var is None:
                arg_var = "$"
            else:
                arg_var = parent._arg_var
        else:
            arg_var = var_prefix

        if parent is None or parent._arg_var is None:
            if isinstance(schema, CMDSchema):
                if not arg_var.endswith("$") and not schema.name.startswith('[') and not schema.name.startswith('{'):
                    arg_var += '.'
                arg_var += f'{schema.name}'.replace('$', '')  # some schema name may contain $
            else:
                raise NotImplementedError()
        else:
            assert isinstance(parent, CMDArgBuilder)
            if isinstance(parent.schema, CMDArraySchemaBase):
                arg_var += '[]'
            elif isinstance(parent.schema, CMDObjectSchemaAdditionalProperties):
                arg_var += '{}'
            elif isinstance(parent.schema, (CMDObjectSchemaBase, CMDObjectSchemaDiscriminator)):
                if not isinstance(schema, CMDObjectSchemaAdditionalProperties):
                    if not arg_var.endswith("$"):
                        arg_var += '.'
                    if isinstance(schema, CMDObjectSchemaDiscriminator):
                        arg_var += schema.get_safe_value()
                    elif isinstance(schema, CMDSchema):
                        arg_var += f'{schema.name}'.replace('$', '')  # some schema name may contain $
                    else:
                        raise NotImplementedError()
            else:
                raise NotImplementedError()
            cls_name = getattr(parent.schema, 'cls', None)
            if cls_name is not None:
                arg_var = arg_var.replace(parent._arg_var, f"@{cls_name}")

        if ref_arg:
            assert ref_args is None

        flatten = None
        sub_ref_args = []
        if not ref_arg and ref_args:
            for arg in ref_args:
                if arg.var == arg_var:
                    ref_arg = arg
                    flatten = False
                    break
                elif arg.var.startswith(f"{arg_var}."):
                    flatten = True  # this argument already flattened
                    sub_ref_args.append(arg)
        sub_ref_args = sub_ref_args or None
        return cls(schema=schema, arg_var=arg_var, ref_arg=ref_arg, sub_ref_args=sub_ref_args, parent=parent, is_update_action=is_update_action, flatten=flatten)

    def __init__(self, schema, arg_var, ref_arg, sub_ref_args, parent=None, is_update_action=False, flatten=None):
        self.schema = schema
        self._parent = parent
        self._arg_var = arg_var
        self._ref_arg = ref_arg
        self._sub_ref_args = sub_ref_args
        self._flatten = flatten
        self._flatten_discriminators = False  # flatten it's discriminators or not
        self._is_update_action = is_update_action

    def get_sub_builder(self, schema, ref_args=None, ref_arg=None):
        return self.new_builder(
            schema=schema, parent=self, ref_args=ref_args, ref_arg=ref_arg, is_update_action=self._is_update_action)

    def _ignore(self):
        if self.schema.frozen:
            return True
        if isinstance(self.schema, CMDSchemaBase):
            assert not self.schema.read_only
            if self.schema.const:
                return True
        return False

    def _build_arg_base(self):
        if self._ignore():
            return None
        arg_cls = self.schema.ARG_TYPE
        assert issubclass(arg_cls, (CMDArgBase, CMDObjectArgAdditionalProperties))
        return arg_cls.build_arg_base(self)

    def _build_arg(self):
        if self._ignore():
            return None

        arg_cls = self.schema.ARG_TYPE
        assert issubclass(arg_cls, CMDArg)
        return arg_cls.build_arg(self)

    def _need_flatten(self):
        if isinstance(self.schema, CMDObjectSchema):
            if self.get_cls():
                # not support to flatten object which is a cls.
                return False
            if self._flatten is not None:
                return self._flatten
            if self.schema.client_flatten:
                return True
            if self.schema.name == "properties" and self.schema.props:
                # flatten 'properties' property by default if it has props
                return True
        if isinstance(self.schema, CMDObjectSchemaDiscriminator):
            return self._parent._flatten_discriminators
        return False

    def get_args(self):
        if self._ignore():
            return []

        arg = self._build_arg()
        assert arg is not None
        if self._need_flatten():
            if isinstance(self.schema, CMDSchema):
                self.schema.arg = None
                if arg.args:
                    for sub_arg in arg.args:
                        if sub_arg.group is None:
                            sub_arg.group = to_camel_case(self.schema.name)
                        if not arg.required:
                            sub_arg.required = False
            return arg.args or []
        elif isinstance(self.schema, CMDSchema):
            self.schema.arg = arg.var
            arg.ref_schema = self.schema

        return [arg, ]

    def get_sub_args(self):
        assert isinstance(self.schema, (CMDObjectSchemaBase, CMDObjectSchemaDiscriminator))
        sub_args = []
        discriminator_mapping = {}
        if self._ref_arg:
            if isinstance(self._ref_arg, CMDClsArgBase):
                # use the linked instance
                unwrapped_ref_arg = self._ref_arg.get_unwrapped()
                assert unwrapped_ref_arg is not None
                sub_ref_args = unwrapped_ref_arg.args
            else:
                sub_ref_args = self._ref_arg.args
        else:
            sub_ref_args = self._sub_ref_args

        if self.schema.discriminators:
            # update self._flatten_discriminators, if any discriminator need flatten, then all discriminator needs to flatten
            for disc in self.schema.discriminators:
                sub_builder = self.get_sub_builder(schema=disc, ref_args=sub_ref_args)
                self._flatten_discriminators = self._flatten_discriminators or sub_builder._need_flatten()
            for disc in self.schema.discriminators:
                sub_builder = self.get_sub_builder(schema=disc, ref_args=sub_ref_args)
                results = sub_builder.get_args()
                sub_args.extend(results)
                if results and not self._flatten_discriminators:
                    assert len(results) == 1
                    if disc.property not in discriminator_mapping:
                        discriminator_mapping[disc.property] = {}
                    discriminator_mapping[disc.property][disc.value] = results[0].var

        if self.schema.props:
            for prop in self.schema.props:
                if prop.name in discriminator_mapping:
                    # If discriminators are not flattened then prop value can be associate with discriminator arguments
                    assert hasattr(prop, 'enum')
                    for item in prop.enum.items:
                        if item.value in discriminator_mapping[prop.name]:
                            item.arg = discriminator_mapping[prop.name][item.value]
                    continue
                sub_builder = self.get_sub_builder(schema=prop, ref_args=sub_ref_args)
                sub_args.extend(sub_builder.get_args())

        if not sub_args:
            return None
        return sub_args

    def get_sub_item(self):
        if hasattr(self.schema, "item") and self.schema.item:
            sub_ref_arg = self._ref_arg.item if self._ref_arg else None
            sub_builder = self.get_sub_builder(schema=self.schema.item, ref_arg=sub_ref_arg)
            return sub_builder._build_arg_base()
        else:
            return None

    def get_any_type(self):
        if hasattr(self.schema, "any_type") and self.schema.any_type and self.get_sub_item() is None:
            return True
        else:
            return False

    def get_additional_props(self):
        if hasattr(self.schema, "additional_props") and self.schema.additional_props:
            sub_ref_arg = self._ref_arg.additional_props if self._ref_arg else None
            sub_builder = self.get_sub_builder(schema=self.schema.additional_props, ref_arg=sub_ref_arg)
            return sub_builder._build_arg_base()
        else:
            return None

    def get_required(self):
        if not self._is_update_action and isinstance(self.schema, CMDSchema):
            return self.schema.required
        return False

    def get_nullable(self):
        if isinstance(self.schema, CMDSchemaBase) and self.schema.nullable:
            return True

        if isinstance(self.schema, CMDSchema):
            # when updated and schema is not required then nullable is true.
            # This can help update command to remove properties
            if not self.schema.required and self._is_update_action:
                return True

        elif isinstance(self.schema, CMDSchemaBase):
            # when updated and the element is nullable
            # This can help update command to remove elements.
            if self._is_update_action:
                return True

        return False

    def get_default(self):
        if self._ref_arg:
            # ref_arg already has default value return it
            if self._ref_arg.default:
                return CMDArgDefault(raw_data=self._ref_arg.default.to_native())
        if self._is_update_action:
            # ignore default for update actions
            return None
        if hasattr(self.schema, 'default') and self.schema.default:
            return CMDArgDefault.build_default(self, self.schema.default)
        return None

    def get_configuration_key(self):
        if self._ref_arg:
            return self._ref_arg.configuration_key
        return None

    def get_prompt(self):
        if self._ref_arg:
            # ref_arg already has prompt return it
            if hasattr(self._ref_arg, "prompt") and self._ref_arg.prompt:
                return self._ref_arg.prompt.__class__(raw_data=self._ref_arg.prompt.to_native())
        if isinstance(self.schema, CMDPasswordSchema):
            return CMDPasswordArgPromptInput(raw_data={"msg": "Password:"})
        return None

    def get_blank(self):
        if self.get_prompt() is not None:
            # disable blank when get prompt is available
            return None

        if self._ref_arg:
            if self._ref_arg.blank:
                return CMDArgBlank(raw_data=self._ref_arg.blank.to_native())
            return None  # ignore the logic from schema

        if isinstance(self.schema, CMDBooleanArgBase):
            blk = CMDArgBlank()
            blk.value = True
            return blk
        return None

    def get_hide(self):
        if self._ref_arg:
            return self._ref_arg.hide  # ignore the logic from schema

        if getattr(self.schema, 'name', None) == 'id' and not self.get_required() and self._parent and \
                isinstance(self.schema, CMDResourceIdSchema):
            if self._arg_var.split('.', maxsplit=1)[-1] == 'id':
                # hide top level 'id' property when it has 'name' property,
                for prop in self._parent.schema.props:
                    if prop.name == 'name':
                        return True
        return False

    def get_var(self):
        return self._arg_var

    @staticmethod
    def _build_option_name(name):
        name = name.replace('_', '-')
        name = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', name)
        name = re.sub('([a-z0-9])([A-Z])', r'\1-\2', name).lower()
        return '-'.join([p for p in name.split('-') if p])

    def get_options(self):
        if self._ref_arg:
            return [*self._ref_arg.options]

        if isinstance(self.schema, CMDObjectSchemaDiscriminator):
            opt_name = self._build_option_name(self.schema.get_safe_value())
        elif isinstance(self.schema, CMDSchema):
            name = self.schema.name.replace('$', '')
            if name == "[Index]" or name == "{Key}":
                assert self._arg_var.endswith(name)
                prefix = self._arg_var[:-len(name)].split('.')[-1]
                prefix = self._inflect_engine.singular_noun(prefix)
                if name == "[Index]":
                    name = f'{prefix}-index'
                elif name == "{Key}":
                    name = f'{prefix}-key'
            elif name.startswith('[].') or name.startswith('{}.'):
                assert self._arg_var.endswith(name)
                prefix = self._arg_var[:-len(name)].split('.')[-1]
                prefix = self._inflect_engine.singular_noun(prefix)
                name = prefix + name[2:]
            name = name.replace('.', '-')
            opt_name = self._build_option_name(name)  # some schema name may contain $
        else:
            raise NotImplementedError()
        return [opt_name, ]

    def get_singular_options(self):
        if self._ref_arg:
            singular_options = getattr(self._ref_arg, 'singular_options', None)
            if singular_options:
                return [*singular_options]

        # Disable singular options by default
        # if isinstance(self.schema, CMDArraySchema):
        #     opt_name = self._build_option_name(self.schema.name.replace('$', ''))  # some schema name may contain $
        #     singular_opt_name = self._inflect_engine.singular_noun(opt_name) or opt_name
        #     if singular_opt_name != opt_name:
        #         return [singular_opt_name, ]
        return None

    def get_help(self):
        if self._ref_arg:
            if self._ref_arg.help:
                return CMDArgumentHelp(raw_data=self._ref_arg.help.to_native())

        if hasattr(self.schema, 'description') and self.schema.description:
            h = CMDArgumentHelp()
            h.short = self.schema.description.replace('\n', ' ')
            return h
        return None

    def get_group(self):
        if self._ref_arg:
            return self._ref_arg.group
        return None

    def get_fmt(self):
        if isinstance(self.schema, CMDObjectSchemaDiscriminator):
            return None
        assert hasattr(self.schema, 'fmt')
        if self.schema.fmt:
            assert isinstance(self.schema.fmt, CMDFormat)
            ref_fmt = getattr(self._ref_arg, 'fmt', None) if self._ref_arg else None
            return self.schema.fmt.build_arg_fmt(self, ref_fmt=ref_fmt)
        return None

    def get_enum(self):
        assert hasattr(self.schema, 'enum')
        if self.schema.enum:
            ref_enum = self._ref_arg.enum if self._ref_arg else None
            enum = CMDArgEnum.build_enum(self.schema.enum, ref_enum=ref_enum)
            return enum
        return None

    def get_cls(self):
        if isinstance(self.schema, CMDObjectSchemaDiscriminator):
            return None
        assert hasattr(self.schema, 'cls')
        return self.schema.cls

    def get_type(self):
        return self.schema._get_type()

    def get_reverse_boolean(self):
        assert isinstance(self.schema, CMDBooleanSchemaBase)
        if self._ref_arg and isinstance(self._ref_arg, CMDBooleanArgBase):
            return self._ref_arg.reverse
        return False

    def get_resource_location_no_rg_default(self):
        assert isinstance(self.schema, CMDResourceLocationSchemaBase)
        if self._ref_arg and isinstance(self._ref_arg, CMDResourceLocationArgBase):
            return self._ref_arg.no_rg_default
        return False



import logging

from schematics.models import Model
from schematics.types import ListType, BaseType, DictType, ModelType, BooleanType, FloatType, IntType, StringType, \
    PolyModelType

from command.model.configuration import CMDSchemaDefault, \
    CMDStringSchema, CMDResourceIdSchema, CMDResourceIdFormat, \
    CMDResourceLocationSchema, \
    CMDObjectSchemaBase, CMDObjectSchemaDiscriminator, CMDObjectSchemaAdditionalProperties, \
    CMDArraySchemaBase, CMDObjectSchema, CMDIdentityObjectSchema, CMDIdentityObjectSchemaBase, CMDClsSchemaBase
from command.model.configuration import CMDSchemaEnum, CMDSchemaEnumItem, CMDSchema, CMDSchemaBase
from swagger.utils import exceptions
from .external_documentation import ExternalDocumentation
from .fields import DataTypeFormatEnum, RegularExpressionField, XmsClientNameField, XmsExternalField, \
    XmsDiscriminatorValueField, XmsClientFlattenField, XmsMutabilityField, XmsClientDefaultField, XNullableField, \
    XmsAzureResourceField, MutabilityEnum, XmsArmIdDetailsField
from .fields import XmsSecretField, XAccessibilityField, XAzSearchDeprecatedField, XSfClientLibField, \
    XApimCodeNillableField, XCommentField, XAbstractField, XADLNameField, XCadlNameField, XTypespecNameField
from .reference import ReferenceField, Linkable
from .x_ms_enum import XmsEnumField
from .xml import XML

logger = logging.getLogger('backend')


def schema_and_reference_schema_claim_function(_, data):
    if isinstance(data, dict):
        if ReferenceSchema._claim_polymorphic(data=data):
            return ReferenceSchema
        else:
            return Schema
    else:
        return None


def _additional_properties_claim_function(_, data):
    if isinstance(data, bool):
        return bool
    elif isinstance(data, dict):
        return schema_and_reference_schema_claim_function(_, data)
    else:
        return None


def _items_claim_function(_, data):
    if isinstance(data, dict):
        return schema_and_reference_schema_claim_function(_, data)
    elif isinstance(data, list):
        if data and ReferenceSchema._claim_polymorphic(data=data[0]):
            return ListType(ModelType(ReferenceSchema))
        else:
            return ListType(ModelType(Schema))
    else:
        return None


class ReferenceSchema(Model, Linkable):
    ref = ReferenceField(required=True)
    description = StringType()
    title = StringType()
    read_only = BooleanType(
        serialized_name="readOnly",
        deserialize_from="readOnly"
    )  # Relevant only for Schema "properties" definitions. Declares the property as "read only". This means that it MAY be sent as part of a response but MUST NOT be sent as part of the request. Properties marked as readOnly being true SHOULD NOT be in the required list of the defined schema. Default value is false.
    type = StringType(
        choices=["array", "boolean", "integer", "number", "object", "string"],
        # https://datatracker.ietf.org/doc/html/draft-zyp-json-schema-04#section-3.5
    )   #

    x_ms_client_name = XmsClientNameField()  # TODO: used for deserialize name

    x_ms_client_flatten = XmsClientFlattenField()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_instance = None
        self.x_ms_azure_resource = False

    def get_disc_parent(self):
        return self.ref_instance.get_disc_parent()

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        self.ref_instance, instance_traces = swagger_loader.load_ref(self.ref, *self.traces, 'ref')
        if isinstance(self.ref_instance, Linkable):
            self.ref_instance.link(swagger_loader, *instance_traces)
        if self.ref_instance.x_ms_azure_resource:
            self.x_ms_azure_resource = True
        if self.ref_instance.read_only:
            # inherit read only from $ref
            self.read_only = True

    def to_cmd(self, builder, support_cls_schema=False, **kwargs):
        model = builder.register_cls_definition(self, support_cls_schema=support_cls_schema, **kwargs)
        if isinstance(model, CMDSchema):
            builder.setup_description(model, self)
            if self.x_ms_client_flatten:
                model.client_flatten = True
        return model

    @classmethod
    def _claim_polymorphic(cls, data):
        return isinstance(data, dict) and "$ref" in data and len(data) <= 7 and \
               set(data.keys()).issubset({"$ref", "description", "title", "readOnly", "type", "x-ms-client-name", "x-ms-client-flatten"})


class Schema(Model, Linkable):
    """
    The Schema Object allows the definition of input and output data types. These types can be objects, but also primitives and arrays. This object is based on the JSON Schema Specification Draft 4 and uses a predefined subset of it. On top of this subset, there are extensions provided by this specification to allow for more complete documentation.
    Further information about the properties can be found in JSON Schema Core and JSON Schema Validation. Unless stated otherwise, the property definitions follow the JSON Schema specification as referenced here.
    """

    ref = ReferenceField()
    format = DataTypeFormatEnum()
    title = StringType()
    description = StringType()
    default = BaseType()

    # Validation keywords for numeric instances (number and integer)
    multiple_of = FloatType(
        min_value=0,
        serialized_name="multipleOf",
        deserialize_from="multipleOf"
    )
    maximum = FloatType()
    exclusive_maximum = BooleanType(
        serialized_name="exclusiveMaximum",
        deserialize_from="exclusiveMaximum",
    )
    minimum = FloatType()
    exclusive_minimum = BooleanType(
        serialized_name="exclusiveMinimum",
        deserialize_from="exclusiveMinimum"
    )

    # Validation keywords for strings
    max_length = IntType(
        min_value=0,
        serialized_name="maxLength",
        deserialize_from="maxLength"
    )
    min_length = IntType(
        min_value=0,
        serialized_name="minLength",
        deserialize_from="minLength"
    )
    pattern = RegularExpressionField()

    # Validation keywords for arrays
    items = PolyModelType(
        [
            ModelType("Schema"),
            ModelType(ReferenceSchema),
            ListType(ModelType("Schema")),
            ListType(ModelType(ReferenceSchema))
        ],
        claim_function=_items_claim_function,
    )
    max_items = IntType(
        min_value=0,
        serialized_name="maxItems",
        deserialize_from="maxItems"
    )
    min_items = IntType(
        min_value=0,
        serialized_name="minItems",
        deserialize_from="minItems"
    )
    unique_items = BooleanType(
        serialized_name="uniqueItems",
        deserialize_from="uniqueItems"
    )

    # Validation keywords for objects
    max_properties = IntType(
        min_value=0,
        serialized_name="maxProperties",
        deserialize_from="maxProperties"
    )
    min_properties = IntType(
        min_value=0,
        serialized_name="minProperties",
        deserialize_from="minProperties"
    )
    required = ListType(StringType(), min_size=1)
    properties = DictType(
        PolyModelType(
            [
                ModelType("Schema"),
                ModelType(ReferenceSchema)
            ],
            claim_function=schema_and_reference_schema_claim_function,
        ),
    )
    additional_properties = PolyModelType(
        [bool, ModelType("Schema"), ModelType(ReferenceSchema)],
        claim_function=_additional_properties_claim_function,
        serialized_name="additionalProperties",
        deserialize_from="additionalProperties"
    )
    discriminator = StringType()  # Adds support for polymorphism. The discriminator is the schema property name that is used to differentiate between other schema that inherit this schema. The property name used MUST be defined at this schema and it MUST be in the required property list. When used, the value MUST be the name of this schema or any schema that inherits it.

    # Validation keywords for any instance type
    enum = ListType(BaseType())
    x_ms_enum = XmsEnumField()
    type = StringType(
        choices=["array", "boolean", "integer", "number", "object", "string"],
        # https://datatracker.ietf.org/doc/html/draft-zyp-json-schema-04#section-3.5
    )
    all_of = ListType(
        PolyModelType(
            [ModelType("Schema"), ModelType(ReferenceSchema)],
            claim_function=schema_and_reference_schema_claim_function,
        ),
        serialized_name="allOf",
        deserialize_from="allOf"
    )

    read_only = BooleanType(
        serialized_name="readOnly",
        deserialize_from="readOnly"
    )  # Relevant only for Schema "properties" definitions. Declares the property as "read only". This means that it MAY be sent as part of a response but MUST NOT be sent as part of the request. Properties marked as readOnly being true SHOULD NOT be in the required list of the defined schema. Default value is false.
    xml = ModelType(XML)  # TODO: # This MAY be used only on properties schemas. It has no effect on root schemas. Adds Additional metadata to describe the XML representation format of this property.
    external_docs = ModelType(
        ExternalDocumentation,
        serialized_name="externalDocs",
        deserialize_from="externalDocs"
    )  # TODO: # Additional external documentation for this schema.
    example = BaseType()  # TODO: # A free-form property to include an example of an instance for this schema.

    x_ms_client_name = XmsClientNameField()  # TODO: used for deserialized name
    x_ms_external = XmsExternalField()  # TODO:
    x_ms_discriminator_value = XmsDiscriminatorValueField()
    x_ms_client_flatten = XmsClientFlattenField()
    x_ms_mutability = XmsMutabilityField()
    x_ms_client_default = XmsClientDefaultField()
    x_ms_arm_id_details = XmsArmIdDetailsField()  # TODO: Add support for it, can be used for resource id template

    x_ms_azure_resource = XmsAzureResourceField()

    x_ms_secret = XmsSecretField()

    x_nullable = XNullableField()  # when true, specifies that null is a valid value for the associated schema

    x_ms_identifiers = ListType(StringType(), serialized_name="x-ms-identifiers", deserialize_from="x-ms-identifiers")

    # specific properties, will not support
    _x_accessibility = XAccessibilityField()  # only used in ContainerRegistry Data plane
    _x_az_search_deprecated = XAzSearchDeprecatedField()  # only used in Search Data Plane
    _x_sf_clientlib = XSfClientLibField()  # only used in ServiceFabric Data Plane and ServiceFabricManagedClusters Mgmt Plane
    _x_apim_code_nillable = XApimCodeNillableField()  # only used in ApiManagement Mgmt Plane
    _x_comment = XCommentField()  # Only used in IoTCenter Mgmt Plane
    _x_abstract = XAbstractField()  # Only used in Logic Mgmt Plane and Web Mgmt Plane
    _x_adl_name = XADLNameField()  # Only used in FluidRelay Mgmt Plane
    _x_enum_Names = ListType(BaseType(), serialized_name="x-enumNames", deserialize_from="x-enumNames")  # Only used in Marketplane Catalog Data Plane
    _x_enum_flags = BooleanType(serialized_name="x-enumFlags", deserialize_from="x-enumFlags")  # Only used in Marketplane Catalog Data Plane
    _x_dictionary_key = ModelType(ReferenceSchema, serialized_name="x-dictionaryKey", deserialize_from="x-dictionaryKey")  # Only used in Marketplane Catalog Data Plane

    _x_cadl_name = XCadlNameField()  # Cadl field name
    _x_typespec_name = XTypespecNameField()  # Typespec field name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_instance = None
        self.disc_parent = None
        self.disc_children = {}

        self.resource_id_templates = set()  # valid when there's only one template

    def get_disc_parent(self):
        assert self.is_linked()
        if self.disc_parent is not None:
            return self.disc_parent
        elif self.discriminator is not None:
            if not self.properties or self.discriminator not in self.properties:
                raise exceptions.InvalidSwaggerValueError(
                    msg="Discriminator property isn't in properties",
                    key=self.traces,
                    value=self.discriminator
                )
            return self
        elif self.ref_instance is not None:
            return self.ref_instance.get_disc_parent()
        return None

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        if self.ref is not None:
            self.ref_instance, instance_traces = swagger_loader.load_ref(self.ref, *self.traces, 'ref')
            if isinstance(self.ref_instance, Linkable):
                self.ref_instance.link(swagger_loader, *instance_traces)
            if self.ref_instance.x_ms_azure_resource:
                self.x_ms_azure_resource = True
            if self.ref_instance.read_only:
                self.read_only = True

        if self.items is not None:
            if isinstance(self.items, list):
                for idx, item in enumerate(self.items):
                    item.link(swagger_loader, *self.traces, 'items', idx)
            else:
                self.items.link(swagger_loader, *self.traces, 'items')

        if self.properties is not None:
            for key, prop in self.properties.items():
                prop.link(swagger_loader, *self.traces, 'properties', key)

        if self.additional_properties is not None and isinstance(self.additional_properties, (Schema, ReferenceSchema)):
            self.additional_properties.link(swagger_loader, *self.traces, 'additionalProperties')

        if self.all_of is not None:
            for idx, item in enumerate(self.all_of):
                item.link(swagger_loader, *self.traces, 'allOf', idx)
                if item.x_ms_azure_resource:
                    self.x_ms_azure_resource = True

        self._link_disc()
        if self.type and self.type != "object" and self.all_of:
            if len(self.all_of) > 1:
                print(f"\tMultiAllOf for {self.type}: {traces}")
            else:
                print(f"\tAllOf for {self.type}: {traces}")

    def _link_disc(self):
        if self.all_of is None:
            return
        for item in self.all_of:
            disc_parent = item.get_disc_parent()
            if disc_parent is not None:
                if self.disc_parent is not None:
                    raise exceptions.InvalidSwaggerValueError(
                        msg="Multiple discriminator parents exists.",
                        key=self.traces, value=None
                    )
                self.disc_parent = disc_parent

                if self.x_ms_discriminator_value is not None:
                    disc_value = self.x_ms_discriminator_value
                elif len(self.traces) > 2 and self.traces[-2] == 'definitions':
                    disc_value = self.traces[-1]  # use the definition name as discriminator value
                else:
                    # Discriminator value is empty. Its not a discriminator
                    logger.warning(f"Discriminator value is empty. : {self.traces}")
                    continue
                if disc_value in self.disc_parent.disc_children:
                    raise exceptions.InvalidSwaggerValueError(
                        msg=f"Duplicated discriminator children for same value",
                        key=self.traces, value=disc_value
                    )
                self.disc_parent.disc_children[disc_value] = self

    def to_cmd(self, builder, **kwargs):
        if self.ref_instance is not None:
            model = builder(self.ref_instance, support_cls_schema=True)
        else:
            model = builder.build_schema(self)

        if isinstance(model, CMDArraySchemaBase):
            if self.all_of is not None:
                # inherit from allOf
                if len(self.all_of) > 1:
                    raise exceptions.InvalidSwaggerValueError(
                        msg=f"Multiple allOf is not supported for `{model.type}` type schema",
                        key=self.traces, value=None
                    )
                model = builder(self.all_of[0], support_cls_schema=True)

            if self.items:
                assert isinstance(self.items, (Schema, ReferenceSchema))
                v = builder(self.items, in_base=True, support_cls_schema=True)
                assert isinstance(v, CMDSchemaBase)
                model.item = v

                # freeze because array item is frozen
                if not model.frozen and model.item.frozen:
                    model.frozen = True

            if self.x_ms_identifiers:
                model.identifiers = []
                for identifier in self.x_ms_identifiers:
                    model.identifiers.append(identifier)

        elif isinstance(model, CMDObjectSchemaBase):
            # props
            prop_dict = {}
            if model.props is not None:
                # inherit from $ref
                for prop in model.props:
                    prop_dict[prop.name] = prop

            if self.all_of:
                # inherit from allOf
                for item in self.all_of:
                    disc_parent = item.get_disc_parent()
                    if disc_parent is not None and builder.find_traces(item.ref_instance.traces):
                        # discriminator parent already in trace, break reference loop
                        continue
                    v = builder(item, in_base=True, support_cls_schema=False)
                    assert isinstance(v, CMDObjectSchemaBase)
                    if v.fmt:
                        model.fmt = v.fmt

                    if v.props:
                        for p in v.props:
                            prop_dict[p.name] = p

                    if disc_parent is not None and not builder.find_traces(disc_parent.traces):
                        # directly use child definition instead of polymorphism.
                        # So the value for discriminator property is const.
                        is_children = False
                        disc_prop = disc_parent.discriminator
                        for disc_value, disc_child in disc_parent.disc_children.items():
                            if disc_child == self:
                                prop_dict[disc_prop].const = True
                                prop_dict[disc_prop].default = CMDSchemaDefault()
                                prop_dict[disc_prop].default.value = disc_value
                                is_children = True
                                break
                        if not is_children and len(self.all_of) == 1 and \
                                model.props is None and model.additional_props is None:
                            # inherit from allOf as reference only
                            model.discriminators = v.discriminators

                    if v.additional_props:
                        model.additional_props = v.additional_props

            if self.properties:
                for name, p in self.properties.items():
                    assert isinstance(p, (Schema, ReferenceSchema))
                    v = builder(p, in_base=False, support_cls_schema=True)
                    if v is None:
                        # ignore by mutability
                        continue
                    assert isinstance(v, CMDSchema)
                    v.name = name
                    prop_dict[name] = v

            if self.required:
                for name in self.required:
                    if name in prop_dict:
                        # read_only property is not required
                        if prop_dict[name].read_only:
                            logger.info(f"Ignore requirement of the read only property: {name}")
                            continue
                        # because required property will not be included in a cls definition,
                        # so it's fine to update it in parent level when prop_dict[name] is a cls definition.
                        prop_dict[name].required = True
                        if MutabilityEnum.Create == builder.mutability:
                            # for create operation
                            # when a property is required, it's frozen status must be consisted with the defined schema.
                            # This can help to for empty object schema.
                            prop_dict[name].frozen = builder.frozen

            # discriminators
            if self.disc_children:
                discriminators = []
                assert self.discriminator is not None
                disc_prop = self.discriminator
                for disc_value, disc_child in self.disc_children.items():
                    if builder.find_traces(disc_child.traces):
                        # discriminator child already in trace, break reference loop
                        continue
                    disc = CMDObjectSchemaDiscriminator()
                    disc.property = disc_prop
                    disc.value = disc_value

                    if disc_prop not in prop_dict:
                        raise exceptions.InvalidSwaggerValueError(
                            msg="Discriminator Property don't exist",
                            key=self.traces,
                            value=[disc_prop, builder.mutability]
                        )
                    if not hasattr(prop_dict[disc_prop], "enum"):
                        raise exceptions.InvalidSwaggerValueError(
                            msg="Invalid Discriminator Property type",
                            key=self.traces,
                            value=[disc_prop, prop_dict[disc_prop].type]
                        )

                    # make sure discriminator value is an enum item of discriminator property
                    if prop_dict[disc_prop].enum is None:
                        prop_dict[disc_prop].enum = CMDSchemaEnum()
                        prop_dict[disc_prop].enum.items = []
                    exist_disc_value = False
                    for enum_item in prop_dict[disc_prop].enum.items:
                        if enum_item.value == disc_value:
                            exist_disc_value = True
                    if not exist_disc_value:
                        enum_item = CMDSchemaEnumItem()
                        enum_item.value = disc_value
                        prop_dict[disc_prop].enum.items.append(enum_item)

                    v = builder(disc_child, in_base=True, support_cls_schema=False)
                    assert isinstance(v, CMDObjectSchemaBase)
                    if v.frozen and prop_dict[disc_prop].frozen:
                        disc.frozen = True
                    if v.props:
                        disc.props = [prop for prop in v.props if prop.name not in prop_dict]
                    if v.discriminators:
                        disc.discriminators = v.discriminators

                    discriminators.append(disc)
                if discriminators:
                    model.discriminators = discriminators

            # convert special properties when self is an azure resource
            if self.x_ms_azure_resource and prop_dict:
                if 'id' in prop_dict and self.resource_id_templates and not prop_dict['id'].frozen:
                    id_prop = prop_dict['id']
                    if not isinstance(id_prop, CMDResourceIdSchema):
                        try:
                            assert isinstance(id_prop, CMDStringSchema)
                        except:
                            raise
                        raw_data = id_prop.to_native()
                        prop_dict['id'] = id_prop = CMDResourceIdSchema(raw_data=raw_data)
                    if len(self.resource_id_templates) == 1:
                        id_prop.fmt = CMDResourceIdFormat()
                        id_prop.fmt.template = [*self.resource_id_templates][0]
                    else:
                        err = exceptions.InvalidSwaggerValueError(
                            msg="Multi resource id templates error",
                            key=self.traces,
                            value=self.resource_id_templates
                        )
                        logger.warning(err)
                if 'location' in prop_dict and not prop_dict['location'].frozen:
                    location_prop = prop_dict['location']
                    if not isinstance(location_prop, CMDResourceLocationSchema):
                        assert isinstance(location_prop, CMDStringSchema)
                        raw_data = location_prop.to_native()
                        prop_dict['location'] = CMDResourceLocationSchema(raw_data=raw_data)

            if prop_dict:
                if "userAssignedIdentities" in prop_dict and "type" in prop_dict:
                    if isinstance(model, CMDObjectSchema):
                        # Transfer to IdentityObjectSchema
                        model = CMDIdentityObjectSchema(model.to_native())
                    elif isinstance(model, CMDObjectSchemaBase):
                        # Transfer to IdentityObjectSchemaBase
                        model = CMDIdentityObjectSchemaBase(model.to_native())
                    else:
                        raise NotImplementedError()
                model.props = []
                for prop in prop_dict.values():
                    model.props.append(prop)

            # additional properties
            if self.additional_properties:
                if isinstance(self.additional_properties, (Schema, ReferenceSchema)):
                    v = builder(self.additional_properties, in_base=True, support_cls_schema=True)
                    if v is not None:
                        assert isinstance(v, CMDSchemaBase)
                        model.additional_props = CMDObjectSchemaAdditionalProperties()
                        model.additional_props.item = v
                        # item is required, it's frozen status must be consisted with the defined schema.
                        # This can help to for empty object.
                        model.additional_props.item.frozen = builder.frozen
                elif isinstance(self.additional_properties, bool) and self.additional_properties is True:
                    # Free-Form Objects
                    # additionalProperties: true
                    model.additional_props = CMDObjectSchemaAdditionalProperties()
                    model.additional_props.any_type = True

            if model.additional_props:
                if builder.read_only:
                    model.additional_props.read_only = builder.read_only
                if builder.frozen:
                    model.additional_props.frozen = builder.frozen

            if self.x_ms_client_flatten and isinstance(model, CMDObjectSchema):
                # client flatten can only be supported for CMDObjectSchema install of CMDObjectSchemaBase.
                # Because CMDObjectSchemaBase will not link with argument
                model.client_flatten = True

            # when all additional_props and props and discriminators of model is frozen then this model is frozen
            if not model.frozen and (model.additional_props or model.props or model.discriminators):  # an empty object is not included
                need_frozen = True
                if model.additional_props:
                    if not model.additional_props.frozen:
                        need_frozen = False
                if model.props:
                    for prop in model.props:
                        if not prop.frozen:
                            need_frozen = False
                            break
                if model.discriminators:
                    for disc in model.discriminators:
                        if not disc.frozen:
                            need_frozen = False
                            break
                # Note: model will always frozen when an unempty object without any props, additional_props or discriminators,
                # If this property is required by parent schema, it will be updated in parent.
                model.frozen = need_frozen

        else:
            if self.all_of is not None:
                # inherit from allOf
                if len(self.all_of) > 1:
                    raise exceptions.InvalidSwaggerValueError(
                        msg=f"Multiple allOf is not supported for `{model.type}` type schema",
                        key=self.traces, value=None
                    )
                model = builder(self.all_of[0], support_cls_schema=True)

        builder.setup_fmt(model, self)
        builder.setup_enum(model, self)
        builder.setup_default(model, self)
        builder.setup_nullable(model, self)

        if isinstance(model, CMDSchema):
            builder.setup_description(model, self)
            builder.setup_secret(model, self)

        return model



from cli.model.common._fields import CLICommandNameField, CLIVersionField, CLICommandConfigurationField
from schematics.models import Model
from schematics.types import ModelType, ListType, StringType, BooleanType
from utils.stage import AAZStageField

from ._help import CLICommandHelp
from ._resource import CLISpecsResource


class CLIAtomicCommandRegisterInfo(Model):
    stage = AAZStageField(required=True)  # the stage used in code, usually it should be consist with command.stage
    confirmation = StringType(min_length=1)

    # TODO: add support for deprecate_info

    class Options:
        serialize_when_none = False


class CLIAtomicCommand(Model):
    names = ListType(field=CLICommandNameField(), min_size=1, required=True)  # full name of a command
    stage = AAZStageField()  # the stage of command
    help = ModelType(CLICommandHelp, required=True)
    register_info = ModelType(
        CLIAtomicCommandRegisterInfo,
        serialized_name="registerInfo",
        deserialize_from="registerInfo"
    )  # register info in command table. If it's not registered in command table, this field is None

    version = CLIVersionField()  # the version of wait command is not required.
    resources = ListType(ModelType(CLISpecsResource), required=True, min_size=1)

    cfg = CLICommandConfigurationField()
    support_no_wait = BooleanType(
        serialized_name="supportNoWait",
        deserialize_from="supportNoWait",
    )

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import ModelType, ListType

from ._command import CMDCommand
from ._fields import CMDCommandGroupNameField


class CMDCommandGroup(Model):
    # properties as tags
    name = CMDCommandGroupNameField(required=True)
    # stage = CMDStageField()

    # properties as nodes
    # help = ModelType(CMDHelp, required=True)
    commands = ListType(ModelType(CMDCommand))  # sub commands
    command_groups = ListType(
        ModelType("CMDCommandGroup"),
        serialized_name='commandGroups',
        deserialize_from='commandGroups',
    )  # sub command groups

    class Options:
        serialize_when_none = False

    def reformat(self, **kwargs):
        if self.command_groups:
            for group in self.command_groups:
                group.reformat(**kwargs)
            self.command_groups = sorted(
                [group for group in self.command_groups if group.commands or group.command_groups],
                key=lambda g: g.name
            )
        if self.commands:
            for command in self.commands:
                command.reformat(**kwargs)
        elif self.command_groups:
            if len(self.command_groups) == 1:
                sub_group = self.command_groups[0]
                self.name += ' ' + sub_group.name
                self.command_groups = sub_group.command_groups
                self.commands = sub_group.commands
                del sub_group

    def link(self):
        if self.commands:
            for command in self.commands:
                command.link()
        if self.command_groups:
            for group in self.command_groups:
                group.link()



import click
import json
import logging
import os
import re
import sys
from flask import Blueprint

from command.controller.specs_manager import AAZSpecsManager
from command.templates import get_templates
from swagger.utils.tools import swagger_resource_path_to_resource_id
from swagger.utils.source import SourceTypeEnum
from utils.config import Config

logger = logging.getLogger('backend')

bp = Blueprint('aaz-cmds', __name__, url_prefix='/AAZ/CMDs', cli_group="command-model")
bp.cli.short_help = "Manage command models in aaz."


def path_type(ctx, param, value):
    import os
    return os.path.expanduser(value)


def resource_id_type(value):
    return swagger_resource_path_to_resource_id(value)


@bp.cli.command("generate-from-swagger", short_help="Generate command models into aaz from swagger specs")
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--swagger-module-path", "--sm",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_MODULE_PATH,
    callback=Config.validate_and_setup_swagger_module_path,
    expose_value=False,
    help="The local path of swagger in module level. It can be substituted for --swagger-path."
)
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--module", '-m',
    default=Config.DEFAULT_SWAGGER_MODULE,
    required=not Config.DEFAULT_SWAGGER_MODULE,
    callback=Config.validate_and_setup_default_swagger_module,
    expose_value=False,
    help="The name of swagger module."
)
@click.option(
    "--resource-provider", "--rp",
    default=Config.DEFAULT_RESOURCE_PROVIDER,
    required=not Config.DEFAULT_RESOURCE_PROVIDER,
    callback=Config.validate_and_setup_default_resource_provider,
    expose_value=False,
    help="The resource provider name."
)
@click.option(
    "--swagger-tag", "--tag",
    required=True,
    help="Swagger tag with input files."
)
@click.option(
    "--workspace-path",
    help="The path to export the workspace for modification."
)
def generate_command_models_from_swagger(swagger_tag, workspace_path=None):
    from swagger.controller.specs_manager import SwaggerSpecsManager
    from command.controller.specs_manager import AAZSpecsManager
    from command.controller.workspace_manager import WorkspaceManager
    from utils.config import Config
    from utils.exceptions import InvalidAPIUsage
    from command.model.configuration import CMDHelp

    try:
        swagger_specs = SwaggerSpecsManager()
        aaz_specs = AAZSpecsManager()

        module_manager = swagger_specs.get_module_manager(Config.DEFAULT_PLANE, Config.DEFAULT_SWAGGER_MODULE)
        rp = module_manager.get_openapi_resource_provider(Config.DEFAULT_RESOURCE_PROVIDER)

        resource_map = rp.get_resource_map_by_tag(swagger_tag)
        if not resource_map:
            raise InvalidAPIUsage(f"Tag `{swagger_tag}` is not exist")

        version_resource_map = {}
        for resource_id, version_map in resource_map.items():
            v_list = [v for v in version_map]
            if len(v_list) > 1:
                raise InvalidAPIUsage(f"Tag `{swagger_tag}` contains multiple api versions of one resource", payload={
                    "Resource": resource_id,
                    "versions": v_list,
                })
            v = v_list[0]
            if v not in version_resource_map:
                version_resource_map[v] = []
            version_resource_map[v].append({
                "id": resource_id
            })

        mod_names = Config.DEFAULT_SWAGGER_MODULE.split('/')
        ws = WorkspaceManager.new(
            name=Config.DEFAULT_SWAGGER_MODULE,
            plane=Config.DEFAULT_PLANE,
            folder=workspace_path or WorkspaceManager.IN_MEMORY,  # if workspace path exist, use workspace else use in memory folder
            mod_names=mod_names,
            resource_provider=rp.name,
            swagger_manager=swagger_specs,
            aaz_manager=aaz_specs,
            source=SourceTypeEnum.OpenAPI,
        )
        for version, resources in version_resource_map.items():
            ws.add_new_resources_by_swagger(
                mod_names=mod_names, version=version, resources=resources
            )

        # provide default short summary
        for node in ws.iter_command_tree_nodes():
            if not node.help:
                node.help = CMDHelp()
            if not node.help.short:
                node.help.short = f"Manage {node.names[-1]}"

        for leaf in ws.iter_command_tree_leaves():
            if not leaf.help:
                leaf.help = CMDHelp()
            if not leaf.help.short:
                n = leaf.names[-1]
                n = n[0].upper() + n[1:]
                leaf.help.short = f"{n} {leaf.names[-2]}"

        if not ws.is_in_memory:
            ws.save()

        ws.generate_to_aaz()

    except InvalidAPIUsage as err:
        logger.error(err)
        sys.exit(1)
    except ValueError as err:
        logger.error(err)
        sys.exit(1)


@bp.cli.command("verify", short_help="Verify data consistency within `aaz` repository.")
@click.option(
    "--aaz-path", "-a",
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="Path of `aaz` repository."
)
def verify():
    def verify_command(file_path, node):
        with open(file_path, "r", encoding="utf-8") as fp:
            content = fp.read()

        paths = re.findall(r"]\(([^)]+)\)", content)
        for path in paths:
            json_path = os.path.join(Config.AAZ_PATH, os.path.splitext(path)[0][1:] + ".json")
            json_path = os.path.normpath(json_path)
            if not os.path.exists(json_path):
                raise Exception(f"{json_path} defined in {file_path} is missing.")

            with open(json_path, "r", encoding="utf-8", errors="ignore") as fp:
                model = json.load(fp)
            group, command = " ".join(node.names[:-1]), node.names[-1]
            for g in model["commandGroups"]:
                if g["name"] == group:
                    if not any(cmd["name"] == command for cmd in g["commands"]):
                        raise Exception(f"There is no {command} command info in {json_path}.")

                    break

            model_set.add(json_path)

        tmpl = get_templates()["command"]
        if not tmpl.render(command=node) == content:
            raise Exception(f"{file_path} cannot be rendered correctly.")

    model_set = set()
    aaz = AAZSpecsManager()
    stack = [(aaz.commands_folder, aaz.tree.root)]  # root nodes

    while stack:
        curr_path, curr_node = stack.pop()
        logger.info(f"Checking {curr_path}")
        if os.path.isdir(curr_path):
            readme_path = os.path.join(curr_path, "readme.md")
            if not os.path.exists(readme_path):
                raise Exception(f"Missing `readme.md` under {curr_path}.")

            with open(readme_path, "r", encoding="utf-8") as fp:
                content = fp.read()

            matches = re.findall(r"## (.+)\n\n(((?!\n##)[\s\S])+)", content)
            for match in matches:
                level = match[0]
                items = re.findall(r"- \[([^[\]]+)]", match[1])

                if level == "Commands":
                    if len(items) != len(set(items)):
                        raise Exception(f"{readme_path} has duplicate command names.")

                    items = set(items)

                    files = {i for i in os.listdir(curr_path) if os.path.isfile(os.path.join(curr_path, i))}
                    files.remove("readme.md")

                    if (cmd_set := set(map(lambda x: x[1:-3], files))) != items:  # _<command_name>.md
                        diff = cmd_set - items or items - cmd_set
                        raise Exception(f"Command info {diff} doesn't match in {readme_path}.")

                    groups = set(curr_node.commands.keys())
                    if groups != items:
                        diff = groups - items or items - groups
                        raise Exception(f"Command info {diff} in tree.json doesn't match in {readme_path}.")

                    for file in files:
                        verify_command(os.path.join(curr_path, file), curr_node.commands[file[1:-3]])
                else:
                    if len(items) != len(set(items)):
                        raise Exception(f"{readme_path} has duplicate command group names.")

                    items = set(items)

                    folders = {i for i in os.listdir(curr_path) if os.path.isdir(os.path.join(curr_path, i))}
                    if folders != items:
                        diff = folders - items or items - folders
                        raise Exception(f"Command group info {diff} doesn't match in {readme_path}.")

                    groups = set(curr_node.command_groups.keys())
                    if groups != set(items):
                        diff = groups - items or items - groups
                        raise Exception(f"Command group info {diff} in tree.json doesn't match in {readme_path}.")

                    for folder in folders:
                        stack.append((os.path.join(curr_path, folder), curr_node.command_groups[folder]))

    for root, dirs, files in os.walk(aaz.resources_folder):
        for file in files:
            if not file.endswith(".json") or file.startswith("client"):  # support data-plane
                continue

            file_path = os.path.join(root, file)
            if file_path not in model_set:
                raise Exception(f"{file_path} is redundant.")



from jinja2.filters import pass_environment
from utils.base64 import b64encode_str
from utils.stage import AAZStageEnum
from utils.plane import PlaneEnum


@pass_environment
def command_group_readme_path(env, names):
    return '/'.join(["", "Commands", *names, 'readme.md'])


@pass_environment
def command_readme_path(env, names):
    return '/'.join(["", "Commands", *names[:-1], f'_{names[-1]}.md'])


@pass_environment
def resource_cfg_path(env, resource):
    plane = resource.plane
    if plane == PlaneEnum.Mgmt:
        plane_folder = plane
    elif PlaneEnum.is_data_plane(plane):
            scope = PlaneEnum.get_data_plane_scope(plane)
            if not scope:
                raise ValueError(f"Invalid plane: Missing scope in data plane '{plane}'")
            plane_folder = '/'.join([PlaneEnum._Data, scope])
    else:
        raise ValueError(f"Invalid plane: '{plane}'")
    return '/'.join(["", "Resources", plane_folder, b64encode_str(resource.id), f"{resource.version}.xml"])


@pass_environment
def stage_label(env, stage, bold=True):
    text = stage
    if stage is None:
        text = AAZStageEnum.Stable
    if bold:
        text = f"**{text}**"
    return text


custom_filters = {
    "command_group_readme_path": command_group_readme_path,
    "command_readme_path": command_readme_path,
    "resource_cfg_path": resource_cfg_path,
    "stage_label": stage_label
}



import ast
import glob
import json
import logging
import os
import pkgutil
import re

from cli.controller.az_profile_generator import AzProfileGenerator
from cli.controller.az_atomic_profile_builder import AzAtomicProfileBuilder
from cli.model.view import CLIModule, CLIViewProfile, CLIViewCommandGroup, CLIViewCommand
from cli.templates import get_templates
from utils import exceptions
from utils.config import Config
from collections import deque

logger = logging.getLogger('backend')


class AzModuleManager:
    _command_group_pattern = re.compile(r'^class\s+(.*)\(.*AAZCommandGroup.*\)\s*:\s*$')
    _command_pattern = re.compile(r'^class\s+(.*)\(.*AAZ(Wait)?Command.*\)\s*:\s*$')
    _is_preview_param = re.compile(r'\s*is_preview\s*=\s*True\s*')
    _is_experimental_param = re.compile(r'\s*is_experimental\s*=\s*True\s*')
    _def_pattern = re.compile(r'^\s*def\s+')
    _aaz_info_pattern = re.compile(r'^\s*_aaz_info\s*=\s*({.*)$')

    @staticmethod
    def pkg_name(mod_name):
        return mod_name.replace('-', '_').lower()

    def get_mod_path(self, mod_name):
        raise NotImplementedError()

    def get_aaz_path(self, mod_name):
        raise NotImplementedError()

    def has_module(self, mod_name):
        mod_folder = self.get_mod_path(mod_name)
        if not os.path.exists(mod_folder):
            #print(f"Invalid module folder: cannot find modules in: '{mod_folder}'")
            return False
        return True

    def find_module_cmd_registered(self, node):
        group_nodes = deque()
        group_nodes.append(node)
        registered_cmds = []

        while len(group_nodes) > 0:
            group_node = group_nodes.popleft()
            if 'command_groups' in group_node and group_node.command_groups is not None:
                for group in group_node.command_groups:
                    group_nodes.append(group_node.command_groups[group])
            if 'commands' in group_node and group_node.commands is not None:
                for command in group_node.commands:
                    cmd_node = group_node.commands[command]
                    if cmd_node.registered:
                        registered_cmds.append(cmd_node.names + [cmd_node.version])
        return registered_cmds

    def load_module(self, mod_name):
        module = CLIModule()
        module.name = mod_name
        module.folder = self.get_mod_path(mod_name)

        module.profiles = {}
        for profile_name in Config.CLI_PROFILES:
            module.profiles[profile_name] = self._load_view_profile(profile_name, self.get_aaz_path(mod_name))
        return module

    def update_module(self, mod_name, profiles, **kwargs):
        aaz_folder = self.get_aaz_path(mod_name)
        generators = {}
        atomic_builder = AzAtomicProfileBuilder(mod_name=mod_name, by_patch=kwargs.pop('by_patch', False))
        for profile_name, profile in profiles.items():
            profile = atomic_builder(profile)
            generators[profile_name] = AzProfileGenerator(aaz_folder, profile)
        for generator in generators.values():
            generator.generate()
        for generator in generators.values():
            generator.save()
        for patch_file, file_data in self._patch_module(mod_name):
            os.makedirs(os.path.dirname(patch_file), exist_ok=True)
            with open(patch_file, 'w', encoding="utf-8") as f:
                f.write(file_data)
        module = CLIModule()
        module.name = mod_name
        module.folder = self.get_mod_path(mod_name)
        module.profiles = profiles
        return module

    _def_load_command_table = re.compile("^(\s+)def\s+load_command_table\(\s*self,\s+(\w+)\s*\):(.*)?$")
    _def_import_load_aaz = re.compile("\s+(import\s+(\w+.)*load_aaz_command_table)\s*$")

    def _patch_module(self, mod_name):
        """Patch the __init__.py file of module"""
        file = os.path.join(os.path.dirname(self.get_aaz_path(mod_name)), '__init__.py')
        if not os.path.exists(file) or not os.path.isfile(file):
            raise exceptions.InvalidAPIUsage(f"Patch Module failed: Cannot find file: {file}")

        with open(file, 'r', encoding="utf-8") as f:
            lines = f.read().split('\n')

        start_line = None
        insert_after = None
        args_name = None
        space = None
        for idx in range(len(lines)):
            line = lines[idx]
            if self._def_import_load_aaz.findall(line):
                # already patched
                logger.debug(f"Module is already patched")
                return
            if start_line is None:
                def_match = self._def_load_command_table.match(line)
                if def_match:
                    start_line = idx
                    insert_after = idx
                    space = def_match[1]
                    args_name = def_match[2]
                    if args_name == '_':
                        args_name = 'args'
                        lines[idx] = "    def load_command_table(self, args):" + def_match[3]
            else:
                if line.startswith(f"{space}{space}return") or line.startswith(f"{space}def ") or not line.startswith(
                        space):
                    # finish the load_command_table function
                    break
                if line.startswith(f"{space}{space}from ") or line.startswith(f"{space}{space}import "):
                    insert_after = idx
        if start_line is None:
            raise exceptions.InvalidAPIUsage(f"Patch Module failed: Cannot find load_command_table function in file: {file}")

        insert_lines = [
            f"{space}{space}from azure.cli.core.aaz import load_aaz_command_table",
            f"{space}{space}try:",
            f"{space}{space}{space}from . import aaz",
            f"{space}{space}except ImportError:",
            f"{space}{space}{space}aaz = None",
            f"{space}{space}if aaz:",
            f"{space}{space}{space}load_aaz_command_table(",
            f"{space}{space}{space}{space}loader=self,",
            f"{space}{space}{space}{space}aaz_pkg_name=aaz.__name__,",
            f"{space}{space}{space}{space}args={args_name}",
            f"{space}{space}{space})"
        ]
        lines = lines[:insert_after + 1] + insert_lines + lines[insert_after + 1:]

        yield file, '\n'.join(lines)

    def _load_view_profile(self, profile_name, aaz_path):
        profile = CLIViewProfile()
        profile.name = profile_name
        profile_folder_name = profile.profile_folder_name
        profile_path = os.path.join(aaz_path, profile_folder_name)
        if not os.path.exists(profile_path):
            return profile
        profile.command_groups = self._load_view_command_groups(path=profile_path)
        return profile

    def _load_view_command_groups(self, *names, path):
        """Load command groups folder in the folder"""
        command_groups = {}
        assert os.path.isdir(path), f'Invalid folder path {path}'
        for name in os.listdir(path):
            sub_path = os.path.join(path, name)
            if not name.startswith('_') and os.path.isdir(sub_path):
                name = name.replace('_', '-')  # transform folder_name to command_group_name
                command_group = self._load_view_command_group(*names, name,
                                                              path=sub_path)  # load command group definition
                if command_group:
                    command_groups[name] = command_group
        if not command_groups:
            return None
        return command_groups

    def _load_view_commands(self, *names, path):
        """Load commands in the folder"""
        commands = {}
        wait_command = None
        assert os.path.isdir(path), f'Invalid folder path {path}'
        for name in os.listdir(path):
            sub_path = os.path.join(path, name)
            if os.path.isfile(sub_path) and name.endswith('.py') and not name.startswith('__') and name.startswith('_'):
                name = name[1:-3].replace('_', '-')  # transform file_name to command_name
                command, is_wait = self._load_view_command(*names, name, path=sub_path)  # load command definition
                if command:
                    if is_wait:
                        wait_command = command
                    else:
                        commands[name] = command
        if not commands:
            return None, None
        return commands, wait_command

    def _load_view_command_group(self, *names, path):
        assert os.path.isdir(path), f'Invalid folder path {path}'
        init_file = os.path.join(path, '__init__.py')
        if not os.path.exists(init_file) or not os.path.isfile(init_file):
            return None
        cmd_group_file = os.path.join(path, '__cmd_group.py')
        if not os.path.exists(cmd_group_file) or not os.path.isfile(cmd_group_file):
            return None

        register_info_lines = None
        find_command_group = False
        with open(cmd_group_file, 'r', encoding="utf-8") as f:
            while f.readable():
                line = f.readline()
                if line.startswith('@register_command_group('):
                    register_info_lines = []
                if self._command_group_pattern.match(line):
                    find_command_group = True
                    break
                if register_info_lines is not None:
                    register_info_lines.append(line)
        if not find_command_group:
            return None

        command_group = CLIViewCommandGroup()
        command_group.names = [*names]

        command_group.command_groups = self._load_view_command_groups(*names, path=path)
        command_group.commands, command_group.wait_command = self._load_view_commands(*names, path=path)
        return command_group

    def _load_view_command(self, *names, path):
        assert os.path.isfile(path), f'Invalid file path {path}'

        register_info_lines = None
        aaz_info_lines = None
        find_command = False
        is_wait_command = False
        with open(path, 'r', encoding="utf-8") as f:
            while f.readable():
                line = f.readline()
                if line.startswith('@register_command('):
                    register_info_lines = []
                _command_pattern_match = self._command_pattern.match(line)
                if _command_pattern_match:
                    find_command = True
                    is_wait_command = _command_pattern_match[2] is not None
                    break
                if register_info_lines is not None:
                    register_info_lines.append(line)
            if not find_command:
                return None, None

            while f.readable():
                line = f.readline()
                if self._def_pattern.findall(line):
                    break
                match = self._aaz_info_pattern.match(line)
                if match:
                    aaz_info_lines = match[1]
                    while f.readable():
                        line = f.readline()
                        line = line.strip()
                        aaz_info_lines += ' ' + line
                        if line.endswith('}'):
                            break
                    break

        if not find_command:
            return None, None

        if not aaz_info_lines:
            raise exceptions.InvalidAPIUsage(f"Command info miss in code: '{' '.join(names)}'")

        try:
            data = ast.literal_eval(aaz_info_lines)
            if 'version' not in data and not is_wait_command:
                logger.info(f"Ignore command without version: '{' '.join(names)}'")
                return None, None
            version_name = data.get('version', None)
        except Exception as err:
            raise exceptions.InvalidAPIUsage(f"Command info invalid in code: '{' '.join(names)}': {err}: {aaz_info_lines}")

        command = CLIViewCommand()
        command.names = [*names]
        command.version = version_name

        if register_info_lines:
            command.registered = True
        return command, is_wait_command

    def _load_view_wait_command(self, *names, path):
        wait_command = None
        assert os.path.isdir(path), f'Invalid folder path {path}'
        for name in os.listdir(path):
            sub_path = os.path.join(path, name)
            if os.path.isfile(sub_path) and name.endswith('.py') and not name.startswith('__') and name.startswith('_'):
                name = name[1:-3].replace('_', '-')  # transform file_name to command_name
                command = self._load_view_command(*names, name, path=sub_path)  # load command definition
        # name = '_wait.py'
        # names = [*cg_names, name]
        # assert os.path.isdir(path), f'Invalid folder path {path}'
        # sub_path = os.path.join(path, name)


class AzMainManager(AzModuleManager):

    @classmethod
    def _find_module_folder(cls):
        cli_folder = Config.CLI_PATH
        if not os.path.exists(cli_folder) or not os.path.isdir(cli_folder):
            raise ValueError(f"Invalid Cli Main Repo folder: '{cli_folder}'")
        module_folder = os.path.join(cli_folder, "src", "azure-cli", "azure", "cli", "command_modules")
        if not os.path.exists(module_folder):
            raise ValueError(f"Invalid Cli Main Repo folder: cannot find modules in: '{module_folder}'")
        return module_folder

    def __init__(self):
        super().__init__()
        module_folder = self._find_module_folder()
        self.folder = module_folder

    def get_mod_path(self, mod_name):
        return os.path.join(self.folder, self.pkg_name(mod_name))

    def get_aaz_path(self, mod_name):
        return os.path.join(self.get_mod_path(mod_name), 'aaz')

    def list_modules(self):
        modules = []
        for _, pkg_name, _ in pkgutil.iter_modules(path=[self.folder]):
            modules.append({
                "name": pkg_name.replace('_', '-'),
                "folder": os.path.join(self.folder, pkg_name)
            })

        return sorted(modules, key=lambda a: a['name'])

    def create_new_mod(self, mod_name):
        mod_path = self.get_mod_path(mod_name)
        templates = get_templates()['main']
        new_files = {}

        # render __init__.py
        file_path = os.path.join(mod_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['__init__.py']
        new_files[file_path] = tmpl.render(
            mod_name=mod_name
        )

        # render _help.py
        file_path = os.path.join(mod_path, '_help.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['_help.py']
        new_files[file_path] = tmpl.render()

        # render _params.py
        file_path = os.path.join(mod_path, '_params.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['_params.py']
        new_files[file_path] = tmpl.render()

        # render commands.py
        file_path = os.path.join(mod_path, 'commands.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['commands.py']
        new_files[file_path] = tmpl.render()

        # render custom.py
        file_path = os.path.join(mod_path, 'custom.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['custom.py']
        new_files[file_path] = tmpl.render()

        # test_folder
        t_path = os.path.join(mod_path, 'tests')
        t_templates = templates['tests']

        # render __init__.py
        file_path = os.path.join(t_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = t_templates['__init__.py']
        new_files[file_path] = tmpl.render()

        profile = Config.CLI_DEFAULT_PROFILE
        tp_path = os.path.join(t_path, profile)
        tp_templates = t_templates['profile']

        # render __init__.py
        file_path = os.path.join(tp_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = tp_templates['__init__.py']
        new_files[file_path] = tmpl.render()

        # render test_*.py
        file_path = os.path.join(tp_path, f'test_{self.pkg_name(mod_name)}.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = tp_templates['test_.py']
        new_files[file_path] = tmpl.render(name=mod_name)

        # written files
        for path, data in new_files.items():
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(data)

        module = CLIModule()
        module.name = mod_name
        module.folder = self.get_mod_path(mod_name)

        module.profiles = {}
        for profile_name in Config.CLI_PROFILES:
            module.profiles[profile_name] = self._load_view_profile(profile_name, self.get_aaz_path(mod_name))

        return module


class AzExtensionManager(AzModuleManager):
    _folder = None
    _folder_is_module = None

    @classmethod
    def _find_module_folder(cls):
        cli_ext_folder = Config.CLI_EXTENSION_PATH
        if not os.path.exists(cli_ext_folder) or not os.path.isdir(cli_ext_folder):
            raise ValueError(f"Invalid Cli Extension Repo folder: '{cli_ext_folder}'")

        if not cls._folder:
            module_paths = []
            for path, _, _ in os.walk(cli_ext_folder):
                if 'azext_' in path:
                    continue
                ext_paths = [path for path in glob.glob(os.path.join(path, 'azext_*')) if os.path.isdir(path)]
                if ext_paths and os.path.exists(os.path.join(path, 'setup.py')):
                    module_paths.append(path)
            module_paths = sorted(module_paths)
            if not module_paths:
                cls._folder = os.path.join(cli_ext_folder, 'src')
                cls._folder_is_module = False
            elif module_paths[0] == cli_ext_folder:
                cls._folder = cli_ext_folder
                cls._folder_is_module = True
            else:
                folders = list(set([os.path.dirname(path) for path in module_paths]))
                if len(folders) > 1:
                    raise ValueError(
                        f"Invalid Cli Extension Repo: '{cli_ext_folder}', Modules in multi folders: {folders}")
                cls._folder = folders[0]
                assert cls._folder.startswith(cli_ext_folder) and len(cls._folder) > len(cli_ext_folder)
                cls._folder_is_module = False
        return cls._folder, cls._folder_is_module

    def __init__(self):
        super().__init__()
        self.folder, self.folder_is_module = self._find_module_folder()

    def get_mod_path(self, mod_name):
        if self.folder_is_module:
            name = self._get_module_name_by_path(self.folder)
            if name == mod_name:
                return self.folder
            else:
                raise exceptions.ResourceConflict(f"Invalid Module name, only support '{name}'")
        else:
            return os.path.join(self.folder, mod_name)

    def _generate_mod_ext_name(self, mod_name):
        return 'azext_' + self.pkg_name(mod_name)

    @staticmethod
    def _get_module_name_by_path(path):
        return os.path.split(path)[-1]

    def get_mod_ext_path(self, mod_name):
        mod_path = self.get_mod_path(mod_name)
        ext_paths = [path for path in glob.glob(os.path.join(mod_path, 'azext_*')) if os.path.isdir(path)]
        if not ext_paths:
            return os.path.join(mod_path, self._generate_mod_ext_name(mod_name))
        else:
            if len(ext_paths) > 1:
                raise exceptions.ResourceConflict(f"Find multi folders start with 'azext_' in module path '{mod_path}'")
            return ext_paths[0]

    def get_aaz_path(self, mod_name):
        return os.path.join(self.get_mod_ext_path(mod_name), 'aaz')

    def list_modules(self):
        modules = []
        if self.folder_is_module:
            modules.append({
                "name": self._get_module_name_by_path(self.folder),
                "folder": self.folder
            })
        else:
            for setup_path in glob.glob(os.path.join(self.folder, '*', 'setup.py')):
                module_folder = os.path.dirname(setup_path)
                modules.append({
                    "name": self._get_module_name_by_path(module_folder),
                    "folder": module_folder
                })
        return modules

    def create_or_update_mod_azext_metadata(self, mod_name):
        from packaging import version
        ext_path = self.get_mod_ext_path(mod_name)
        metadata_path = os.path.join(ext_path, 'azext_metadata.json')
        if os.path.exists(metadata_path):
            with open(metadata_path, 'r', encoding="utf-8") as f:
                metadata = json.load(f)
            if "azext.minCliCoreVersion" not in metadata or \
                    version.parse(metadata["azext.minCliCoreVersion"]) < Config.MIN_CLI_CORE_VERSION:
                metadata["azext.minCliCoreVersion"] = str(Config.MIN_CLI_CORE_VERSION)
        else:
            metadata = {
                "azext.isPreview": True,
                "azext.minCliCoreVersion": str(Config.MIN_CLI_CORE_VERSION),
            }
        return metadata_path, metadata

    def update_module(self, mod_name, profiles, **kwargs):
        module = super().update_module(mod_name, profiles, **kwargs)
        file_path, ext_metadata = self.create_or_update_mod_azext_metadata(mod_name)
        with open(file_path, 'w', encoding="utf-8") as f:
            f.write(json.dumps(ext_metadata, indent=4, sort_keys=True))
        return module

    def create_new_mod(self, mod_name):
        if self.folder_is_module:
            raise exceptions.ResourceConflict(
                f"Cannot create a new module in cli extension repo('{Config.CLI_EXTENSION_PATH}'), "
                f"because the repo is an extension module"
            )
        mod_path = self.get_mod_path(mod_name)
        templates = get_templates()['extension']
        new_files = {}

        # render HISTORY.rst
        file_path = os.path.join(mod_path, 'HISTORY.rst')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['HISTORY.rst']
        new_files[file_path] = tmpl.render()

        # render README.md
        file_path = os.path.join(mod_path, 'README.md')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['README.md']
        new_files[file_path] = tmpl.render(
            mod_name=mod_name
        )

        # render setup.cfg
        file_path = os.path.join(mod_path, 'setup.cfg')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['setup.cfg']
        new_files[file_path] = tmpl.render()

        # render setup.py
        file_path = os.path.join(mod_path, 'setup.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = templates['setup.py']
        new_files[file_path] = tmpl.render(
            mod_name=mod_name
        )

        # azext_* folder
        ext_path = self.get_mod_ext_path(mod_name)
        ext_templates = templates['azext_']

        # render __init__.py
        file_path = os.path.join(ext_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = ext_templates['__init__.py']
        new_files[file_path] = tmpl.render(
            mod_name=mod_name
        )

        # render _help.py
        file_path = os.path.join(ext_path, '_help.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = ext_templates['_help.py']
        new_files[file_path] = tmpl.render()

        # render _params.py
        file_path = os.path.join(ext_path, '_params.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = ext_templates['_params.py']
        new_files[file_path] = tmpl.render()

        # render commands.py
        file_path = os.path.join(ext_path, 'commands.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = ext_templates['commands.py']
        new_files[file_path] = tmpl.render()

        # render custom.py
        file_path = os.path.join(ext_path, 'custom.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = ext_templates['custom.py']
        new_files[file_path] = tmpl.render()

        # azext_metadata.json
        file_path = os.path.join(ext_path, 'azext_metadata.json')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        file_path, ext_metadata = self.create_or_update_mod_azext_metadata(mod_name)
        new_files[file_path] = json.dumps(ext_metadata, indent=4, sort_keys=True)

        # test_folder
        t_path = os.path.join(ext_path, 'tests')
        t_templates = ext_templates['tests']

        # render __init__.py
        file_path = os.path.join(t_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = t_templates['__init__.py']
        new_files[file_path] = tmpl.render()

        profile = Config.CLI_DEFAULT_PROFILE
        tp_path = os.path.join(t_path, profile)
        tp_templates = t_templates['profile']

        # render __init__.py
        file_path = os.path.join(tp_path, '__init__.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = tp_templates['__init__.py']
        new_files[file_path] = tmpl.render()

        # render test_*.py
        file_path = os.path.join(tp_path, f'test_{self.pkg_name(mod_name)}.py')
        if os.path.exists(file_path):
            raise exceptions.ResourceConflict(f"File already exist: '{file_path}'")
        tmpl = tp_templates['test_.py']
        new_files[file_path] = tmpl.render(name=mod_name)

        # written files
        for path, data in new_files.items():
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(data)

        module = CLIModule()
        module.name = mod_name
        module.folder = self.get_mod_path(mod_name)

        module.profiles = {}
        for profile_name in Config.CLI_PROFILES:
            module.profiles[profile_name] = self._load_view_profile(profile_name, self.get_aaz_path(mod_name))

        return module
